/*
 * Copyright (c) 1999 Marc Haber <Marc.Haber@gmx.de>
 * All rights reserved.
 *
 * This software's redistribution and use in source and binary forms
 * is defined in the GNU General Public License.
 */

#undef DEBUG
//#define DEBUG 1

#include <iostream.h>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <algo.h>
#include <regex.h>
#include <time.h>

const string MyName = "ipnm_cache: ";

#ifdef DEBUG
#define ITEMSEP "\n"
#define DEBCERR(x) cerr << MyName << x << "\n"
#else
#define ITEMSEP "; "
#define DEBCERR(x)
#endif



int ieof;

class RouteInfo
{
public:
  string Target;
  string Prefix;
};


class Ifacedata
{
public:
  string IP;
  string Prefix;
  string Device;
  string Broadcast;
};

map<string,Ifacedata> Ifaces;
map<string,vector<RouteInfo> > Targets;
set<string> Exports;

string ReadLine()
{
  string Line;
	
  if (!cin)
  {
    ieof=1;
  }
  else
  {
    getline(cin,Line);
  }
  #ifdef DEBUG
  cerr << "Read: " << Line << ":\n";
  #endif
  return Line;
}

void ReplaceString( string &In, const string Search, const string Target)
{
  // #ifdef DEBUG
  // cerr << "ReplaceString(" << In << "," << Search << "," << Target << ");\n";
  // cerr << "In.find(Search)==" << In.find(Search) << ":\n";
  //   // cerr << "Search.length()==" << Search.length() << ":\n";
  //#endif

  string::size_type pos;
  
  while( (pos=In.find(Search)) != string::npos )
  {
    In.replace(pos,Search.length(),Target);
  }

  //#ifdef DEBUG
  //cerr << "Output: " << In << ":\n";
  //#endif
}

string GetEnv(const string name)
{
  
  char *pEnv = getenv(name.c_str());
  if( pEnv == NULL )
  {
	return "";
  }
  else
  {
    string Env = pEnv;
    ReplaceString(Env,"{","");
    ReplaceString(Env,"}","");
    ReplaceString(Env,",","");
    return Env;
  }
}


string Trim(string s)
{
  string::iterator i=s.begin();
  
  while(isspace(*i++) && i!=s.end());
  
  if( i-- != s.begin() ) 
  {
    s.erase(s.begin(),i); 
  }
  
  
  i=s.end();
  while(isspace(*--i) && i!=s.begin());
  if( i++ != s.end() ) 
  {
	 s.erase(i,s.end()); 
  }
  
  
  if( s == " " ) s="";
  return s;
}


int CheckAge()
/* This function reads the first line from input that is supposed to
   contain the creation time of the cache data. If it is more than n
   minutes in the past, an error message is printed to stderr and the
   program terminates
   ##+#FIXME: Automatically recreate cache file if it is too old */
{
  const time_t MaxAge = 10*60;
  string Line = ReadLine();
  
  string::size_type i = Line.find(" ");
  
  time_t CreationTime = atoi( Line.substr(0,i).c_str() );
  time_t Now = time(NULL);
  time_t Age = Now-CreationTime;
  
#ifdef DEBUG
  cerr << MyName << "Input Data Time:   " << CreationTime << "\n";
  cerr << MyName << "Current Time:      " << Now << "\n";
  cerr << MyName << "Age of Input Data: " << Age << "\n";
#endif

#ifndef DEBUG
  if( CreationTime == 0 )
  {	    
    cerr << MyName << "Input data creation time is 0. Age check disabled\n";
  }
  else if( Age > MaxAge )
  {
    cerr << MyName << "Input data too old.\n"
		 << MyName << "create new data with\n"
		 << MyName << "(date +'%s %a %b %e %H:%M:%S %Z %Y'; ip addr; ip route ) > ifacedata.cache\n"
		 << MyName << "and retry\n";


	exit(1);
  }
#endif  
}

string ExtractMatch( string Input, regmatch_t Match )
{
  /* extracts matches substring */
  #ifdef DEBUG1
  cerr << MyName << "ExtractMatch\n";
  cerr << MyName << "Input       = " << Input << "\n";
  cerr << MyName << "Match.rm_so = " << Match.rm_so << "\n";
  cerr << MyName << "Match.rm_eo = " << Match.rm_eo << "\n";
  #endif

  if( Match.rm_so == -1 && Match.rm_eo == -1 )
  {
    // a non-match
    return "";
  }
  
  return Input.substr(Match.rm_so,Match.rm_eo-Match.rm_so);
}


int ParseInputData()
/* This function parses ifconfig output into the appropriate
   data structures */
{
enum ParseState { ProcessIfaces,
                  ProcessRoutingTable, End, Error };

regex_t reCompiled;
const int nMatch = 15;
regmatch_t Match[nMatch];

string Ifacename;
string Realname, IP, Bcast, Prefix;

string Line = ReadLine();
ParseState State = ProcessIfaces;

ieof = 0;

regex_t reRoutingEntry;
regcomp( &reRoutingEntry, "^(([0-9]{1,3}\\.){3}[0-9]{1,3}(/[0-9]{1,2})?)|(default)", REG_EXTENDED);
	
while( !ieof && State!=End)
{
  if( State == ProcessIfaces )
  {
    #ifdef DEBUG
    cerr << MyName << "ProcessIfaces\n";
    #endif
    if( regexec( &reRoutingEntry, Line.c_str(), 0, Match , 0 ) == 0 )
    {
      State = ProcessRoutingTable;
      #ifdef DEBUG
      cerr << MyName << "State: ProcessRoutingTable\n";
      #endif
    }
    else
    {
      regcomp( &reCompiled, "^[[:space:]]{4}inet[[:space:]](([0-9]{1,3}\\.){3}[0-9]{1,3})(/[0-9]{1,2})[[:space:]](brd[[:space:]](([0-9]{1,3}\\.){3}[0-9]{1,3})[[:space:]]+)?.*[[:space:]]([a-zA-Z0-9]{1,15})(\\:([a-zA-Z0-9]{1,15}))?$", REG_EXTENDED|REG_ICASE);

      /* Match 1: IP address
       * Match 2: reserved
       * Match 3: Prefix
       * Match 4: reserved
       * Match 5: broadcast address
       * Match 6: reserved
       * Match 7: hardware interface name
       * Match 8: reserved
       * Match 9: IP address name          */
	    
      if( regexec( &reCompiled, Line.c_str(), 10, Match, 0) == 0)
      {
        #ifdef DEBUG
        for( int i=0; i<=9; i++ )
        {
	  cerr << MyName << "Match[" << i << "] = " << ExtractMatch( Line, Match[i] ) << ";\n";
	}
     	#endif

	string IP       =ExtractMatch(Line, Match[1]);
	string Prefix   =ExtractMatch(Line, Match[3]);
	string Bcast    =ExtractMatch(Line, Match[5]);
        string Device   =ExtractMatch(Line, Match[7]);
	string Label    =ExtractMatch(Line, Match[9]);

       	if( Label == "" )
	{
	  Label = Device;
	}
	  	      
        DEBCERR( "Label: " << Label );
        #ifdef DEBUG
        cerr << MyName << "Device: " << Device << ":\n";
       	cerr << MyName << "IP: "     << IP << ":\n";
       	cerr << MyName << "Bcast: "  << Bcast << ":\n";
       	cerr << MyName << "Prefix: " << Prefix << ":\n";
        #endif

	Ifaces[Label].IP        = IP;
       	Ifaces[Label].Broadcast = Bcast;
       	Ifaces[Label].Prefix    = Prefix;
       	Ifaces[Label].Device    = Device;
      }
      else
      {
	#ifdef DEBUG
	cerr << MyName << "no IP line: " << Line << "\n";
	#endif
      }
      regfree( &reCompiled );
    }
  }
  
  if( State == ProcessRoutingTable )
  {
    #ifdef DEBUG
    cerr << MyName << "ProcessRoutingTable\n";
    #endif
    if( Line == "" )
	{
	  State = End;
	}
	else
	{
          #ifdef DEBUG
          cerr << MyName << "Match: " << Line << ":\n";
          #endif
	  regcomp( &reCompiled,
"^((([0-9]{1,3}\\.){3}[0-9]{1,3})(/[0-9]{1,2})?|(default))[[:space:]](via[[:space:]](([0-9]{1,3}\\.){3}[0-9]{1,3})[[:space:]])?dev[[:space:]]([a-zA-Z0-9:]{1,16})(.*)$", REG_EXTENDED|REG_ICASE);


	  /* Match 1: target IP address/prefix or "default"
           * Match 2: target IP address (if not "default")
	   * Match 3: reserved
           * Match 4: Prefix (if not "default")
           * Match 5: "default" (if "default")
	   * Match 6: reserved
	   * Match 7: gateway ("via" clause)
	   * Match 8: reserved
           * Match 9: outgoing interface ("dev" clause)
	   * Match 10: rest of line */
		
	  if( regexec( &reCompiled, Line.c_str(), 11, Match, 0) == 0 )
	  {
	    #ifdef DEBUG
	    for( int i=0; i<=10; i++ )
	    {
	      cerr << MyName << "Match[" << i << "] = " << ExtractMatch( Line, Match[i] ) << ";\n";
	    }
     	    #endif
		  
	    string Target = ExtractMatch( Line, Match[1] );
	    string Prefix;
	    #ifdef DEBUG
	    cerr << MyName << "Target:  " << Target << ":\n";
	    #endif
	    if( Target == "default" )
	    {
	      Target = "0.0.0.0";
	      Prefix = "/0";
	    }
	    else
	    {
	      Target = ExtractMatch( Line, Match[2] );
	      Prefix = ExtractMatch( Line, Match[4] );
	    }
	    string Device = ExtractMatch( Line, Match[9] );
		  
	    #ifdef DEBUG
	    cerr << MyName << "Target:  " << Target << ":\n";
	    cerr << MyName << "Prefix:  " << Prefix << ":\n";
	    cerr << MyName << "Device:  " << Device << ":\n";
	    #endif
	    RouteInfo Route;
	    Route.Target  = Target;
	    Route.Prefix  = Prefix;
	    Targets[Device].push_back(Route);
	  }
	  else
	  {
	    State = Error;
	  }
	}
  }
    
  if( State == Error )
  {
    cerr << MyName << "State Error reached.\n";
	exit(1);
  }
  else
  {
    Line = ReadLine();
  }
}

}


string BuildBlist(set<string> in)
{
  string out="{";
  string sep = " ";
  set<string>::const_iterator Ii;
  for( Ii = in.begin(); Ii != in.end(); ++Ii)
  {
	out+=sep+*Ii;
	sep=" , ";
  }
  out+=" }";
  return out;
}

void ExpandPlusNotation( set<string> &List )
{
  // expand interface names given in + notation
  
  set<string>::const_iterator Ilist;
  for( Ilist = List.begin(); Ilist != List.end(); ++Ilist)
  {
    if ( "+" == Ilist->substr(Ilist->length()-1,1) )
    {
      #ifdef DEBUG
      cerr << "begin wildcard expansion for " << *Ilist << "\n";
      #endif
      string mask = *Ilist;
      mask.replace(mask.length()-1,1,"");
      #ifdef DEBUG
      cerr << "searching for interfaces beginning with " << mask << "\n";
      #endif
      List.erase(*Ilist);
      map<string,Ifacedata>::const_iterator Iif;
      for( Iif = Ifaces.begin(); Iif != Ifaces.end(); ++Iif)
      {
	const string &Label=Iif->first;
	const Ifacedata &If=Iif->second;
        if( If.Device.substr(0,mask.length()) == mask )
	{
	  List.insert(If.Device);
	}
      }
    }
  }
	
}

void ExpandDeviceToLabel(  set<string> &List )
{
  // expand device names to interface labels

  set<string>OutList;
	
  set<string>::const_iterator Ilist;
  for( Ilist = List.begin(); Ilist != List.end(); ++Ilist)
  {
    map<string,Ifacedata>::const_iterator Iif;
    for( Iif = Ifaces.begin(); Iif != Ifaces.end(); ++Iif)
    {
      const string &Label=Iif->first;
      const Ifacedata &If=Iif->second;
      if( If.Device == *Ilist )
      {
        OutList.insert(Label);
      }
    }
  }
  
  List = OutList;
}

void OutputWantedData(set<string> &IncludeList, set<string> &ExcludeList)
{
  // set<string> oIPNM;
  set<string> oIP;
  // set<string> oBCNM;
  set<string> oBC;
  set<string> oTarget;
  set<string> oDevice;
  set<string> oLabel;

  #ifdef DEBUG
  cerr << MyName << "List output from input\n";
  cerr << MyName << "IncludeList: ";
  copy(IncludeList.begin(), IncludeList.end(),
       ostream_iterator<string>(cerr, " ") );
  cerr << endl;
  cerr << MyName << "ExcludeList: ";
  copy(ExcludeList.begin(), ExcludeList.end(),
       ostream_iterator<string>(cerr, " ") );   
  cerr << endl;
  #endif

  ExpandPlusNotation(IncludeList);
  ExpandPlusNotation(ExcludeList);

  #ifdef DEBUG
  cerr << MyName << "List output after ExpandPlusNotationt\n";
  cerr << MyName << "IncludeList: ";
  copy(IncludeList.begin(), IncludeList.end(),
       ostream_iterator<string>(cerr, " ") );
  cerr << endl;
  cerr << MyName << "ExcludeList: ";
  copy(ExcludeList.begin(), ExcludeList.end(),
       ostream_iterator<string>(cerr, " ") );   
  cerr << endl;
  #endif

  ExpandDeviceToLabel(IncludeList);
  ExpandDeviceToLabel(ExcludeList);
  
  #ifdef DEBUG
  cerr << MyName << "List output after ExpandDeviceToLabel\n";
  cerr << MyName << "IncludeList: ";
  copy(IncludeList.begin(), IncludeList.end(),
       ostream_iterator<string>(cerr, " ") );
  cerr << endl;
  cerr << MyName << "ExcludeList: ";
  copy(ExcludeList.begin(), ExcludeList.end(),
       ostream_iterator<string>(cerr, " ") );   
  cerr << endl;
  #endif

  set<string> List;

  set_difference(IncludeList.begin(), IncludeList.end(),
		 ExcludeList.begin(), ExcludeList.end(),
                 inserter(List, List.begin()) );

  #ifdef DEBUG
  cerr << MyName << "List: ";
  copy(List.begin(), List.end(),
       ostream_iterator<string>(cerr, " ") );   
  cerr << endl;
  #endif
	
  set<string>::const_iterator Ilist;
  for( Ilist = List.begin(); Ilist != List.end(); ++Ilist )
  {
	Ifacedata &If=Ifaces[*Ilist];
	#ifdef DEBUG
	cerr << MyName << "Output Data for " << *Ilist << ":\n";
	cerr << MyName << "IPAddress/Prefix: " << If.IP+If.Prefix << ":\n";
	#endif
	oLabel.insert(*Ilist);
        // oIPNM.insert(If.IPAddress+If.Prefix);
	oIP.insert(If.IP);
	// oBCNM.insert(If.Broadcast+IP.Prefix);
	oBC.insert(If.Broadcast);
	oDevice.insert(If.Device);
	vector<RouteInfo>::const_iterator Ti;
	for( Ti = Targets[If.Device].begin(); Ti != Targets[If.Device].end(); ++Ti)
	{
	  #ifdef DEBUG
	  cerr << MyName << "Route: " << Ti->Target+Ti->Prefix << ":\n";
	  #endif
	  oTarget.insert(Ti->Target+Ti->Prefix);
	}
  }

  
  #ifdef DEBUG
  cerr << MyName << "Begin OutputWantedData\n";
  #endif
  
  cout //<< "IPNMOFIF=\""   << BuildBlist(oIPNM)      << "\"" << ITEMSEP
       << "IF_IP=\""    << BuildBlist(oIP)        << "\"" << ITEMSEP
       // << "BCNMOFIF=\""   << BuildBlist(oBCNM)      << "\"" << ITEMSEP
       << "IF_BC=\""    << BuildBlist(oBC)        << "\"" << ITEMSEP
       << "IF_TRG=\""   << BuildBlist(oTarget)    << "\"" << ITEMSEP
       << "IF_LABEL=\"" << BuildBlist(oLabel)     << "\"" << ITEMSEP
       << "IF_DEV=\""   << BuildBlist(oDevice)    << "\"" << ITEMSEP;
  // Exports.insert("IPNMOFIF");
  Exports.insert("IF_IP");
  // Exports.insert("BCNMOFIF");
  Exports.insert("IF_BC");
  Exports.insert("IF_TRG");
  Exports.insert("IF_LABEL");
  Exports.insert("IF_DEV");
}

void OutputAllData()
{
  #ifdef DEBUG
  cerr << MyName << "Begin OutputAllData\n";
  #endif
  map<string,Ifacedata>::const_iterator Iif;
  set<string> Labels;

  set<string> DefaultLabels;
	
  for( Iif = Ifaces.begin(); Iif != Ifaces.end(); ++Iif)
  {
	const string     &Label=Iif->first;
	const Ifacedata  &If   =Iif->second;
	Labels.insert(Label);
	set<string> Target;
	vector<RouteInfo>::const_iterator Ti;
	for( Ti = Targets[If.Device].begin(); Ti != Targets[If.Device].end(); ++Ti)
	{
	  string thisTarget = Ti->Target+Ti->Prefix;
	  #ifdef DEBUG
	  cerr << MyName << "Route: " << thisTarget << ":\n";
	  #endif
	  Target.insert(thisTarget);
	  if( thisTarget == "0.0.0.0/0" )
	  {
	    #ifdef DEBUG
	    cerr << MyName << "default route\n";
	    #endif
	    DefaultLabels.insert(Label);
	  }
	}
    cout << "IF_" << Label << "_IPNM=\"{ "  << If.IP+If.Prefix         << " }\"" << ITEMSEP
         << "IF_" << Label << "_IP=\"{ "    << If.IP                   << " }\"" << ITEMSEP
         << "IF_" << Label << "_BCNM=\"{ "  << If.Broadcast+If.Prefix  << " }\"" << ITEMSEP
         << "IF_" << Label << "_BC=\"{ "    << If.Broadcast            << " }\"" << ITEMSEP
         << "IF_" << Label << "_DEV=\"{ "   << If.Device               << " }\"" << ITEMSEP
	 << "IF_" << Label << "_LABEL=\"{ " << Label                   << " }\"" << ITEMSEP
         << "IF_" << Label << "_TRG=\""     << BuildBlist(Target)      << "\"" << ITEMSEP;
	Exports.insert("IF_"+Label+"_IPNM");
	Exports.insert("IF_"+Label+"_IP");
	Exports.insert("IF_"+Label+"_BCNM");
	Exports.insert("IF_"+Label+"_BC");
	Exports.insert("IF_"+Label+"_DEV");
	Exports.insert("IF_"+Label+"_LABEL");
	Exports.insert("IF_"+Label+"_TRG");
  }
  cout << "IF_LABELS=\"" << BuildBlist(Labels) << "\"" << ITEMSEP;
  Exports.insert("IF_LABELS");
  cout << "IF_DEFAULT=\"" << BuildBlist(DefaultLabels) << "\"" << ITEMSEP;
  Exports.insert("IF_DEFAULT");
  #ifdef DEBUG
  cerr << MyName << "End OutputAllData\n";
  #endif
}

vector<string> SpaceSplit( string Input )
/* Split <Input> at spaces. Return vector */
{
  vector<string> Out;
  string This;
  Input = Trim(Input);
  while( Input != "" )
  {
    Input+=" ";
    int FirstSpace=Input.find_first_of(' ');
    DEBCERR( "Input: "      << Input );
    DEBCERR( "FirstSpace: " << FirstSpace );
    This=Input.substr(0,FirstSpace);
    Input.replace(0,FirstSpace+1,"");
    Input=Trim(Input);
    This=Trim(This);
    Out.push_back(This);
  }
  return Out;
}


void OutputClassTargets()
{
  vector<string> ClassList = SpaceSplit(GetEnv("IF_CLASSES"));
  map<string,vector<string> > Classes;

  // Find class membership from environment

  for( vector<string>::const_iterator Ci=ClassList.begin(); Ci!=ClassList.end(); Ci++ )
  {
    Classes[*Ci]=SpaceSplit(GetEnv(*Ci));
  }
       
  // build list of all targets for all classes and output that data
  
  set<string> TargetSet, EmptySet;
  for( vector<string>::const_iterator Ci=ClassList.begin(); Ci!=ClassList.end(); Ci++)
  {
    DEBCERR( "searching targets for " << *Ci );
    map<string,Ifacedata>::const_iterator Iif;
  
    for( Iif = Ifaces.begin(); Iif != Ifaces.end(); ++Iif)
	{
	  DEBCERR( "is " << Iif->first << " member of " << *Ci << "?" );
	  
	  vector<string>::const_iterator Inti;
          for( Inti = Classes[*Ci].begin(); Inti != Classes[*Ci].end(); ++Inti)
	  {
	    DEBCERR( *Inti );
	    if( *Inti == Iif->first )
	    {
	      DEBCERR( "yes" );
	      // interface is member of class
              vector<RouteInfo>::const_iterator Ti;
	      for( Ti = Targets[Iif->second.Device].begin(); Ti != Targets[Iif->second.Device].end(); ++Ti)
	      {
	        TargetSet.insert(Ti->Target+Ti->Prefix);
	      }
	    }
	    else
	    {
	      DEBCERR( "no" );
	    }
	  }
	}
	
	cout << "TARGETS_" << *Ci << "=\"" << BuildBlist(TargetSet) << "\"" << ITEMSEP;
	Exports.insert("IF_"+*Ci+"_TRG");
	TargetSet = EmptySet;
  }
}



void OutputExports(set<string> Set)
{
  string sep = "";
  set<string>::const_iterator Ii;
  for( Ii = Set.begin(); Ii != Set.end(); ++Ii)
  {
	cout << sep << "export " << *Ii;
	sep=ITEMSEP;
  }
}


int main(int argc, char* argv[])
{
  set<string> IncludeIfaces;
  set<string> ExcludeIfaces;
  
  #ifdef DEBUG
  cerr << MyName << "argc: " << argc << "\n";
  #endif
  
  CheckAge();
  ParseInputData();

  int i;
  int neg=0;
  for( i=1; i<argc; ++i)
  {
	string Arg = argv[i];
	if( Arg=="-" )
        {
	  neg=1;
	  continue;
	}
	
	if( Arg.substr(0,1) == "-" )
	{
	  neg=1;
	  Arg.erase(0,1);
	}
		
	if( Arg=="," || Arg=="{" || Arg=="}" || Arg=="" )
	{
	  continue;
	}
	
	if( neg )
	{
	  #ifdef DEBUG
	  cerr << MyName << "Adding Interface " << Arg << ": to Exclude List\n";
	  #endif

	  ExcludeIfaces.insert(Arg);
	  neg=0;
	}
    else
	{
	  #ifdef DEBUG
	  cerr << MyName << "Adding Interface " << Arg << ": to Include List\n";
	  #endif
	  IncludeIfaces.insert(Arg);
	}
  }

  OutputWantedData(IncludeIfaces, ExcludeIfaces);
//  OutputAllData();
//  OutputClassTargets();
  OutputExports(Exports);
  
  #ifdef DEBUG
  cerr << "\n";
  #endif
	
  return 0;
}
